# tag: table rendering base class

package HTML::Tabulate;

use 5.005;
use Carp;
use Data::Dumper;
use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK);

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw();
@EXPORT_OK = qw(&render);

$VERSION = '0.11';
my %VALID_ARG = (
    table => 'HASH/SCALAR',
    tr => 'HASH', 
    thtr => 'HASH',
    th => 'HASH',
    td => 'HASH',
    fields => 'ARRAY',
    in_fields => 'ARRAY',
    labels => 'SCALAR/HASH',
    label_links => 'HASH',
    stripe => 'ARRAY/SCALAR',
    null => 'SCALAR',
    trim => 'SCALAR',
    style => 'SCALAR',
#   limit => 'SCALAR',
#   output => 'SCALAR',
#   first => 'SCALAR',
#   last => 'SCALAR',
    field_attr => 'HASH',
);
my %VALID_FIELDS = (
    -defaults => 'HASH',
);
my %FIELD_ATTR = (
    escape => 'SCALAR',
    value => 'SCALAR/CODE/ARRAY',
    format => 'SCALAR/CODE',
    link => 'SCALAR/CODE',
    label => 'SCALAR/ARRAY',
    label_link => 'SCALAR/CODE',
);

# -------------------------------------------------------------------------
# Provided for subclassing
sub get_valid_arg
{
    return wantarray ? %VALID_ARG : \%VALID_ARG;
}

# Provided for subclassing
sub get_valid_fields
{
    return wantarray ? %VALID_FIELDS : \%VALID_FIELDS;
}

# Provided for subclassing
sub get_field_attributes
{
    return wantarray ? %FIELD_ATTR : \%FIELD_ATTR;
}

#
# Check $self->{defn} for invalid arguments or types
#
sub check_valid
{
    my ($self, $defn) = @_;

    # Check top-level args
    my %valid = $self->get_valid_arg();
    my (@invalid, @badtype);
    for (sort keys %$defn) {
        push @invalid, $_ if ! exists $valid{$_};
        my $type = ref $defn->{$_};
        push @badtype, $_ 
            if $type && $type ne 'SCALAR' && $valid{$_} !~ m/$type/;
        push @badtype, $_ 
            if ! $type && $valid{$_} !~ m/SCALAR/;
    }
    croak "[check_valid] invalid argument found: " . join(',',@invalid) 
        if @invalid;
    croak "[check_valid] invalid types for argument: " . join(',',@badtype)
        if @badtype;

    # Check special fields
    %valid = $self->get_valid_fields();
    @invalid = ();
    @badtype = ();
    for (sort grep(/^-/, keys(%{$defn->{field_attr}})) ) {
        push @invalid, $_ if ! exists $valid{$_};
        my $type = ref $defn->{field_attr}->{$_};
        push @badtype, $_ 
            if $type && $type ne 'SCALAR' && $valid{$_} !~ m/$type/;
        push @badtype, $_ 
            if ! $type && $valid{$_} !~ m/SCALAR/;
    }
    croak "[check_valid] invalid field argument found: " . join(',',@invalid) 
        if @invalid;
    croak "[check_valid] invalid types for field argument: " . join(',',@badtype)
        if @badtype;

    # Check field attributes
    $self->{field_attr} ||= $self->get_field_attributes();
    %valid = %{$self->{field_attr}};
    @badtype = ();
    for my $field (keys %{$defn->{field_attr}}) {
        croak "[check_valid] invalid field argument entry '$field': " . 
            $defn->{field_attr}->{$field} 
                if ref $defn->{field_attr}->{$field} ne 'HASH';
        for (sort keys %{$defn->{field_attr}->{$field}}) {
            next if ! exists $valid{$_};
            next if ! $valid{$_};
            my $type = ref $defn->{field_attr}->{$field}->{$_};
            if (! ref $valid{$_}) {
                push @badtype, $_ 
                    if $type && $type ne 'SCALAR' && $valid{$_} !~ m/$type/;
                push @badtype, $_ 
                    if ! $type && $valid{$_} !~ m/SCALAR/;
            }
            elsif (ref $valid{$_} eq 'ARRAY') {
                if ($type) {
                    push @badtype, $_;
                }
                else {
                    my $val = $defn->{field_attr}->{$field}->{$_};
                    push @badtype, "$_ ($val)" if ! grep /^$val$/, @{$valid{$_}};
                }
            }
            else {
                croak "[check_valid] invalid field attribute entry for '$_': " . 
                    ref $valid{$_};
            }
        }
        croak "[check_valid] invalid type for '$field' field attribute: " . 
            join(',',@badtype) if @badtype;
    }
}

#
# Merge $hash1 and $hash2 together, returning the result. Performs a 
#   shallow (one-level deep) hash merge unless the field is defined in 
#   the @recurse_keys array, in which case we do a full recursive merge.
#
sub merge
{
    my $self = shift;
    my $hash1 = shift || {};
    my $hash2 = shift;

    croak "[merge] invalid hash1 '$hash1'" if ref $hash1 ne 'HASH';
    croak "[merge] invalid hash2 '$hash2'" if $hash2 && ref $hash2 ne 'HASH';

    my $single_arg = ! $hash2;

    # Use $self->{defn} as $hash1 if only one argument
    if ($single_arg) {
        $hash2 = $hash1;
        $hash1 = $self->{defn};
    }

    # Check hash2 for valid args (except when recursive)
    my $sub = (caller(1))[3];
    $self->check_valid($hash2) if substr($sub, -7) ne '::merge';

    my %merge;
    # Add hash1 to %merge
    @merge{ keys %$hash1 } = values %$hash1;

    # Add hash2 to %merge 
    my @recurse_keys = qw(field_attr);
    for my $key (keys %$hash2) {
        # If this value is a hashref on both sides, do a shallow hash merge
        #   unless we need to do a proper recursive merge
        if (ref $hash2->{$key} eq 'HASH' && ref $merge{$key} eq 'HASH') {
            # Recursive merge
            if (grep /^$key$/, @recurse_keys) {
                $merge{$key} = $self->merge($hash1->{$key}, $hash2->{$key});
            }
            # Shallow hash merge
            else {
                @{$merge{$key}}{ keys %{$hash1->{$key}}, keys %{$hash2->{$key}} } = (values %{$hash1->{$key}}, values %{$hash2->{$key}});
            }
        }
        # Otherwise (scalars, arrayrefs etc) just copy the value
        else {
            $merge{$key} = $hash2->{$key};
        }
    }

    # In void context update $self->{defn}
    if (! defined wantarray) {
        $self->{defn} = \%merge;
        # Must invalidate derived $self->{def2} when $self->{defn} changes
        delete $self->{def2} if exists $self->{def2};
    }
    else {
        return \%merge 
    }
}

# Initialisation
sub init
{
    my $self = shift;
    my $defn = shift || {};
    croak "[init] invalid defn '$defn'" if $defn && ref $defn ne 'HASH';

    # Map $defn table => 1 to table => {} for cleaner merging
    $defn->{table} = {} if $defn->{table} && ! ref $defn->{table};

    # Initialise $self->{defn} by merging defaults and $defn
    $self->{defn} = $self->merge({
        style => 'down', 
        table => {},
        field_attr => { -defaults => {}, },
    }, $defn);

    return $self;
}

sub new 
{
    my $class = shift;
    my $self = {};
    bless $self, $class;
    $self->init(@_);
}

# -------------------------------------------------------------------------
#
# If deriving field names, also derive labels (if not already defined)
#
sub derive_labels
{
    my ($self) = shift;
    my $def2 = $self->{def2};
    return unless ref $def2 eq 'HASH';
    my $fields = $self->{def2}->{fields};
    return unless ref $fields eq 'ARRAY';

    $def2->{field_attr} ||= {};
    for (@$fields) {
        $def2->{field_attr}->{$_} ||= {};
        if (! $def2->{field_attr}->{$_}->{label}) {
            my $label = $_;
            $label =~ s/_+/ /g;
            $label = join ' ', map { ucfirst($_) } split(/\s+/, $label);
            $def2->{field_attr}->{$_}->{label} = $label;
        }
    }
}

#
# Try and derive a reasonable field list from $self->{def2} using the set data.
#   Croaks on failure.
#
sub derive_fields
{
    my ($self, $set) = @_;
    my $defn = $self->{def2};

    # For DBIx::Recordset iterator, prefetch the first row and use its keys
    if (ref $set eq 'DBIx::Recordset') {
        require DBIx::Recordset;
        my $row = $set->First();
        $self->{prefetch} = $row;
        $defn->{fields} = [ sort keys %$row ] if ref $row;
    }
    # For arrays
    elsif (ref $set eq 'ARRAY') {
        if (! @$set) {
            $defn->{fields} = [];
            return;
        }
        my $obj = $set->[0];
        # Arrayref of hashrefs
        if (ref $obj eq 'HASH') {
            $defn->{fields} = [ sort keys %$obj ];
        }
        # Arrayref of arrayrefs - access via subscripts unless labels are defined
        elsif (ref $obj eq 'ARRAY') {
            if ($defn->{labels}) {
                croak "[derive_fields] no fields found and cannot derive fields from data arrayrefs";
            }
            # Arrayref of arrayrefs, labels off
            else {
                $defn->{fields} = [ 0 .. $#$obj ];
            }
        }
        # For Class::DBI objects, derive via columns groups
        elsif ($obj->isa('Class::DBI')) {
            my @col = $obj->columns('Tabulate');
            @col = ( $obj->columns('Essential'), $obj->columns('Others') ) 
                if ! @col && $obj->columns('Essential');
            @col = $obj->columns('All') if ! @col;
            $defn->{fields} = [ @col ] if @col;
        }
        # If all else fails, try treating as a hash
        unless (ref $defn->{fields} && @{$defn->{fields}}) {
            if (! defined eval { $defn->{fields} = [ sort keys %$obj ] }) {
                croak "[derive_fields] no fields found and initial object '$obj' is strange type";
            }
        }
    }
    # Else looks like a single object - check for Class::DBI
    elsif (ref $set && ref $set ne 'HASH' && $set->isa('Class::DBI')) {
        my @col = $set->columns('Tabulate');
        @col = ( $set->columns('Essential'), $set->columns('Others') ) 
            if ! @col && $set->columns('Essential');
        @col = $set->columns('All') if ! @col;
        $defn->{fields} = [ @col ] if @col;
    }
    # Otherwise try treating as a hash
    elsif (defined eval { keys %$set }) {
        my $first = (sort keys %$set)[0];
        # Check whether first value is reference
        if (my $ref = ref $set->{$first}) {
            # Hashref of hashrefs
            if ($ref eq 'HASH') {
                $defn->{fields} = [ sort keys %{$set->{$first}} ];
            }
            elsif (ref $set->[0] ne 'ARRAY') {
                croak "[derive_fields] no fields found and first row '" . $set->[0] . "' is strange type";
            }
            # Hashref of arrayrefs - fatal only if labels => 1
            elsif ($defn->{labels}) {
                croak "[derive_fields] no fields found and cannot derive fields from data arrayrefs";
            }
            # Hashref of arrayrefs, labels off
            else {
                $defn->{fields} = [ 0 .. $#{$set->[$first]} ];
            }
        }
        else {
            $defn->{fields} = [ sort keys %$set ];
        }
    }
    else {
        croak "[derive_fields] no fields found and set '$set' is strange type: $@";
    }
}

#
# Deep copy routine, swiped from a Randal Schwartz column
#
sub deepcopy 
{
    my $this = shift;
    if (! ref $this) {
        return $this;
    } elsif (ref $this eq "ARRAY") {
        return [map deepcopy($_), @$this];
    } elsif (ref $this eq "HASH") {
        return {map { $_ => deepcopy($this->{$_}) } keys %$this};
    } elsif (ref $this eq "CODE") {
        return $this;
    } else {
        die "what type is $_? (" . ref($this) . ")";
    }
}

#
# Create a derived table definition (def2) by doing a set of one-off or 
#   dataset-specific mappings on the current table definition e.g. deriving 
#   a field list if none is set, setting up a field map for arrayref-of-
#   arrayref sets, and mapping top-level shortcuts into their field 
#   attribute equivalents.
#
sub munge_defn
{
    my ($self, $set) = @_;

    # Use $self->{def1}, if that exists, else $self->{defn} as base
    my $defn = $self->{def1} || $self->{defn};

    # If already done, return unless we require any dataset-specific mappings
    if ($self->{def2}) {
        return unless 
            ref $defn->{fields} ne 'ARRAY' || 
            ! @{$defn->{fields}} ||
            (ref $set eq 'ARRAY' && @$set && ref $set->[0] eq 'ARRAY');
    }

    # Copy $defn to $self->{def2}
    $self->{def2} = deepcopy($defn);
    my $def2 = $self->{def2};

    # Try to derive field list if not set
    $self->derive_fields($set) 
        if ref $def2->{fields} ne 'ARRAY' || ! @{$def2->{fields}};
    # Derive field labels
    $self->derive_labels();

    # Set up a field map for arrayref-of-arrayref sets
    my $pos = 0;
    my $fields = ref $def2->{in_fields} eq 'ARRAY' ? $def2->{in_fields} : $def2->{fields};
    $def2->{field_map} = { map  { $_ => $pos++; } @$fields }
        if ref $set eq 'ARRAY' && @$set && ref $set->[0] eq 'ARRAY';

    # Map top-level 'labels' and 'label_links' hashrefs into fields
    if (ref $def2->{labels} eq 'HASH') {
        for (keys %{$def2->{labels}}) {
            $def2->{field_attr}->{$_} ||= {};
            $def2->{field_attr}->{$_}->{label} = $def2->{labels}->{$_};
        }
    }
    if (ref $def2->{label_links} eq 'HASH') {
        for (keys %{$def2->{label_links}}) {
            $def2->{field_attr}->{$_} ||= {};
            $def2->{field_attr}->{$_}->{label_link} = $def2->{label_links}->{$_};
        }
    }

    # Map top-level 'th' hashref into -defaults label_ attributes
    if (ref $def2->{th} eq 'HASH') {
        for (keys %{$def2->{th}}) {
            $def2->{field_attr}->{-defaults}->{"label_$_"} = $def2->{th}->{$_}
                if ! exists $def2->{field_attr}->{-defaults}->{"label_$_"};
        }
    }
    # Map top-level 'td' hashref into -defaults
    if (ref $def2->{td} eq 'HASH') {
        $def2->{field_attr}->{-defaults} = { %{$def2->{td}}, %{$def2->{field_attr}->{-defaults}} };
    }
}

# -------------------------------------------------------------------------
#
# Return the given HTML $tag with attributes from the $attr hashref.
#   An attribute with a non-empty value (i.e. not '' or undef) is rendered
#   attr="value"; one with a value of '' is rendered as a 'bare' attribute
#   (i.e. no '='); one with undef is ignored (allowing unset CGI parameters
#   to be ignored).
#
sub start_tag
{
    my ($self, $tag, $attr) = @_;
    my $str = "<$tag";
    if (ref $attr eq 'HASH') {
        for my $a (sort keys %$attr) {
            if ($attr->{$a} ne '') {
                $str .= qq( $a="$attr->{$a}");
            }
            elsif (defined $attr->{$a}) {
                $str .= qq( $a);
            }
        }
    }
    $str .= ">";
    return $str;
}

sub end_tag
{
    my ($self, $tag) = @_;
    return "</$tag>";
}

# Provided for subclassing
sub start_table 
{
    my ($self) = @_;
    return '' if exists $self->{def2}->{table} && ! $self->{def2}->{table};
    return $self->start_tag('table',$self->{def2}->{table}) . "\n";
}

# Provided for subclassing
sub end_table 
{
    my ($self) = @_;
    return '' if exists $self->{def2}->{table} && ! $self->{def2}->{table};
    return $self->end_tag('table') . "\n";
}

#
# Format the given data item using formatting field attributes (e.g. format, 
#   link, escape etc.)
#
sub cell_format
{
    my ($self, $data, $fattr, $row, $field) = @_;
    my $data_in = $data;

    # 'format' subroutine or sprintf pattern
    if ($fattr->{format}) {
        my $ref = ref $fattr->{format};
        croak "[cell_format] invalid '$field' format: $ref"
            if $ref && $ref ne 'CODE';
        $data = &{$fattr->{format}}($data, $row || {}) if $ref eq 'CODE';
        $data = sprintf $fattr->{format}, $data if ! $ref;
    }

    # 'escape' boolean for simple tag escaping (defaults to on)
    if ($fattr->{escape} || ! exists $fattr->{escape}) {
        $data =~ s/</&lt;/g;
        $data =~ s/>/&gt;/g;
    }

    # 'link' subroutine or sprintf pattern
    if ($fattr->{link}) {
        my $ldata;
        my $ref = ref $fattr->{link};
        croak "[cell_format] invalid '$field' link: $ref"
            if $ref && $ref ne 'CODE';
        $ldata = &{$fattr->{link}}($data_in, $row || {}) if $ref eq 'CODE';
        $ldata = sprintf $fattr->{link}, $data_in if ! $ref;
        $data = qq(<a href="$ldata">$data</a>);
    }

    return $data;
}

sub label
{
    my ($self, $field) = @_;

    # Use first label if arrayref
    if (ref $self->{def2}->{field_attr}->{$field}->{label} eq 'ARRAY') {
        return $self->{def2}->{field_attr}->{$field}->{label}->[0];
    }
    else {
        return $self->{def2}->{field_attr}->{$field}->{label} || ucfirst($field);
    }
}

#
# Add in any extra (conditional) defaults for this field. 
# Provided for subclassing.
#
sub cell_merge_extras
{
    return ();
}

#
# Merge default and field attributes once each for labels and data
#
sub cell_merge_defaults
{
    my ($self, $row, $field) = @_;
    $self->{def2}->{field_attr}->{$field} ||= {};
    my $fattr = { %{$self->{def2}->{field_attr}->{-defaults}},
                  $self->cell_merge_extras($row, $field),
                  %{$self->{def2}->{field_attr}->{$field}}, };

    # For labels, merge and defaults + field attributes, keeping only label_ ones
    if (! defined $row) {
        my $fattr2 = {};
        # Copy and convert label_ attributes into $fattr
        for (keys %$fattr) {
            next if substr($_,0,6) ne 'label_';
            $fattr2->{substr($_,6)} = $fattr->{$_};
        }
        # Set 'value' from 'label'
        $fattr2->{value} = $self->label($field);
        $fattr = $fattr2;
    }

    # For data, merge defaults and the field attributes, discarding label_ ones
    else {
        # Remove all label_ attributes
        for (keys %$fattr) { delete $fattr->{$_} if substr($_,0,6) eq 'label_'; }
    }

    # Create td_attr by removing all $fattr attributes in $field_attr
    my %td_attr = %$fattr;
    for (keys %td_attr) { delete $td_attr{$_} if exists $self->{field_attr}->{$_} }

    # If data, save for subsequent rows
    if ($row) {
        $fattr->{td_attr} = \%td_attr;
        $self->{def2}->{field_attr}->{$field} = $fattr;
    }

    return ($fattr, \%td_attr);
}

#
# Set and format the data for a single (data) cell or item
#
sub cell_content
{
    my ($self, $row, $field, $fattr) = @_;
    $fattr ||= $self->{def2}->{field_attr}->{$field};

    # Set data from $row
    my $data;
    if (ref $row eq 'ARRAY') {
        my $i = keys %{$self->{def2}->{field_map}} ? $self->{def2}->{field_map}->{$field} : $field;
        $data = $row->[ $i ];
    }
    elsif (ref $row) {
        $data = $row->{$field};
    }
    # 'value' literal or subref takes precedence over row
    if ($fattr->{value}) {
        my $ref = ref $fattr->{value};
        if (! $ref) {
            # $data = sprintf $fattr->{value}, $data;
            $data = $fattr->{value};
        }
        elsif ($ref eq 'CODE') {
            $data = &{$fattr->{value}}($data, $row);
        }
        elsif ($ref eq 'ARRAY') {
            $data = $fattr->{value}->[0];
        }
        else {
            croak "[cell_content] invalid '$field' value: $ref";
        };
    }

    # Format
    $data =~ s/^\s*(.*?)\s*$/$1/ if $self->{def2}->{trim};
    $data = $self->cell_format($data, $fattr, $row, $field) if $data != '';
    $data = $self->{def2}->{null} if defined $self->{def2}->{null} && $data eq '';

    return $data;
}

#
# Wrap cell in <th> or <td> table tags
#
sub cell_tags 
{
    my ($self, $data, $row, $field, $tx_attr) = @_;

    my $tag = ! defined $row ? 'th' : 'td';
    return $self->start_tag($tag, $tx_attr) . $data . $self->end_tag($tag);
}

#
# Render a single table cell or item
#
sub cell 
{
    my ($self, $row, $field) = @_;
    my ($fattr, $tx_attr);

    # Merge default and field attributes first time through (labels + data)
    if (! defined $row || ! $self->{def2}->{field_attr}->{$field}->{td_attr}) {
        ($fattr, $tx_attr) = $self->cell_merge_defaults($row, $field);
    }
    else {
        $tx_attr = $self->{def2}->{field_attr}->{$field}->{td_attr};
    }

    # Generate output
    my $out = $self->cell_content($row, $field, $fattr);
    return $self->cell_tags($out, $row, $field, $tx_attr);
}

#
# Modify the $h hashref with a bgcolor attribute for every nth row.
#   $stripe is either an n-ary arrayref of colour values or a scalar 
#   (a scalar is used as a binary stripe)
#
sub stripe
{
    my ($self, $h, $rownum, $stripe) = @_;
    return if $stripe eq '' || ref $h ne 'HASH';
             
    # Treat a scalar stripe as the second of a binary stripe
    $stripe = [ undef, $stripe ] if ! ref $stripe;
                     
    my $r = int($rownum % scalar(@$stripe)) - 1;
    if ($stripe->[$r]) {
        $h->{bgcolor} = $stripe->[$r]
            unless $rownum == 0 && exists $h->{bgcolor};
    }
    else {
        delete $h->{bgcolor} unless $rownum == 0;
    }
}

#
# Render a single table row (style 'down')
#
sub row_down 
{
    my ($self, $row, $rownum) = @_;
    my $defn = $self->{def2};

    # Table striping
    $defn->{tr} = {} if ref $defn->{tr} ne 'HASH';
    my $tr;
    if ($rownum == 0) {
        $tr = $defn->{thtr} || $defn->{tr};
    }
    else {
        $tr = $defn->{tr};
    }
    $self->stripe($tr, $rownum, $defn->{stripe}) if $defn->{stripe};

    # Render row into $out
    my $out = '';
    $out .= $self->start_tag('tr', $tr);
    for my $f (@{$defn->{fields}}) {
        # Labels/headings
        if ($rownum == 0) {
            $out .= $self->cell(undef, $f);
        }
        # Data
        else {
            $out .= $self->cell($row, $f);
        }
    }
    $out .= $self->end_tag('tr');
    $out .= "\n";
}

#
# Render the table body with successive records down the page
#
sub body_down 
{
    my ($self, $set) = @_;
    my $body = '';

    # Labels/headings
    my @fields = @{$self->{def2}->{fields}} 
        if ref $self->{def2}->{fields} eq 'ARRAY';
    if ($self->{def2}->{labels} && @fields) {
        $body .= $self->row_down(undef, 0);
    }

    # Data
    my $r = 1;
    my $row = $self->{prefetch};
    if (ref $set eq 'DBIx::Recordset') {
        require DBIx::Recordset;
        $row ||= $set->First();
        do {
            $body .= $self->row_down($row, $r);
            $r++;
        } while ($row = $set->Next());
    }
    elsif (ref $set eq 'ARRAY') {
        for my $row (@$set) {
            $body .= $self->row_down($row, $r);
            $r++;
        }
    }
    elsif (ref $set eq 'HASH' || eval { keys %$set }) {
        # Check first value - drill down further unless non-reference
        my $k = $fields[0] || (sort keys %$set)[0];
        if (! ref $set->{$k}) {
            $body .= $self->row_down($set, $r);
            $r++;
        }
        else {
            for $k (sort keys %$set) {
                $body .= $self->row_down($set->{$k}, $r);
                $r++;
            }
        }
    }
    else {
        croak "invalid Tabulate data type '$set'";
    }

    return $body;
}

#
# Render a single table row (style 'across')
#
sub row_across
{
    my ($self, $data, $rownum, $field) = @_;
    my $tr = $self->{def2}->{tr} || {};

    # Begin row
    $self->stripe($tr, $rownum, $self->{def2}->{stripe}) 
        if $self->{def2}->{stripe};
    my $body = $self->start_tag('tr', $tr);

    # Label/heading
    $body .= $self->cell(undef, $field) if $self->{def2}->{labels};

    # Data
    for my $row (@$data) {
        $body .= $self->cell($row, $field);
    }

    # End row
    $rownum++;
    $body .= $self->end_tag('tr', $tr) . "\n";
}

#
# Render the table body with successive records across the page 
#   (i.e. fields down the page)
#
sub body_across 
{
    my ($self, $set) = @_;

    # Fetch the full data set
    my @data = ();
    if (ref $set eq 'DBIx::Recordset') {
        require DBIx::Recordset;
        my $row = $set->First();
        if (ref $row) {
            do {
                push @data, $row;
            }
            while ($row = $set->Next());
        }
    }
    elsif (ref $set eq 'ARRAY') {
        @data = @$set;
    }
    elsif (ref $set eq 'HASH' || eval { keys %$set }) {
        @data = ( $set );
    }
    else {
        croak "invalid Tabulate data type '$set'";
    }

    # Merge any thtr (which isn't doable with 'across') into th
#   $self->{def2}->{th} = $self->merge($self->{def2}->{thtr}, $self->{def2}->{th} || {})
#       if $self->{def2}->{thtr};

    # Iterate over fields
    my $rownum = 1;
    my $body = '';
    for my $field (@{$self->{def2}->{fields}}) {
        $body .= $self->row_across(\@data, $rownum, $field);
    }

    return $body;
}

# -------------------------------------------------------------------------
#
# Render the data set $set using the settings in $self->{defn} + $defn,
#   returning the resulting string.
#
sub render
{
    my ($self, $set, $defn) = @_;
    croak "[render] invalid set '$set'" if ! ref $set;

    # If $self not blessed, this is a procedural call, $self is $defn
    $self = HTML::Tabulate->new($self) 
        if ref $self eq 'HASH' || ref $self eq 'ARRAY';

    # Merge any $defn here, but for this render only
    $self->{def1} = $self->merge($self->{defn}, $defn) if ref $defn eq 'HASH';
    delete $self->{def2};

    # Ensure the definition set has been appropriately munged
    $self->munge_defn($set);

    # Start table
    my $out = $self->start_table();

    # Style-specific bodies (default is 'down')
    if ($self->{def2}->{style} eq 'down') {
        $out .= $self->body_down($set);
    }
    elsif ($self->{def2}->{style} eq 'across') {
        $out .= $self->body_across($set);
    }
    else {
        croak "[render] invalid style '$self->{def2}->{style}'";
    }

    # End table
    $out .= $self->end_table();

    return $out;
}

1;

__END__

=head1 NAME

HTML::Tabulate - HTML table rendering class


=head1 SYNOPSIS

    use HTML::Tabulate;

    # Setup some initial defaults
    $t = HTML::Tabulate->new({ 
        table => { border => 0, cellpadding => 0, cellspacing => 3 },
        th => { class => 'foobar' },
        null => '&nbsp;',
        labels => 1,
    });

    # Setup some data-specific settings
    $table = {
        fields => [ qw(emp_id name title) ],
        field_attr => {
            emp_id => {
                format => '%-05d',
                link => "emp.html?id=%s",
                align => 'right',
            },
            name => {
                format => sub { uc(shift) },
            }
        },
    };

    # Render the data
    print $t->render($data, $table);


    
=head1 DESCRIPTION

HTML::Tabulate is used to render/display a given set of data in an 
HTML table. It takes a data set and a presentation definition and 
applies the presentation to the data set to produce the HTML table 
output. The presentation definition accepts arguments corresponding 
to HTML table tags ('table', 'tr', 'th', 'td' etc.), to define 
attributes for those tags, plus additional arguments for other 
aspects of the presentation. HTML::Tabulate supports advanced 
features like automatic striping, arbitrary cell formatting, 
link creation, etc.

Supported data sets include arrayrefs of arrayrefs (as produced 
by the DBI selectall_arrayref, for example), arrayrefs of hashrefs, 
simple hashref (for a single row table), or iterator objects that
support first() and next() methods (like DBIx::Recordset objects or 
Class::DBI iterators). 

By default arrayref-based datasets are interpreted as containing 
table rows; a column-based interpretation can be forced using 
style => 'across'.



=head1 AUTHOR

Gavin Carr <gavin@openfusion.com.au>


=head1 COPYRIGHT

Copyright 2003, Gavin Carr. All Rights Reserved.

This program is free software. You may copy or redistribute it under the 
same terms as perl itself.

=cut
